/* This file adds some functions to the global PortalTools object.  The functions
 * are large so they are being placed in their own file for readability, in a self
 * executing anonymous function that adds to the pre-existing global object.
 * This should run before require.js has a chance to do anything, but require.js,
 * or other code, trying to run between the definition of PortalTools and this
 * file is a potential bug to watch for.
 *
 * Also, this is intentionally not done using require.js.  Since JS modules will
 * eventually become standard, I don't want a third party library to act as a 
 * gate keeper of our own code for downstream developers.
 */
;(function() {
	"use strict";
	
	var getOptionsObject = function() {
		var options = {};
		options.skip = 0;
		options.top = 10;
		options.orderBy = null;
		options.orderAsc = true;
		options.filter = [];
		options.sourceStore = null;
		options.attributeInfoOnly = false;
		options.addFilter = function(attribute, operator, value) {
			options.filter.push({attribute: attribute, operator: operator, value: value});
		};
		return options;
	}
	
	// url, string
	// filterAttributeMap, instances of AttributeMap
	var CollectionDataSource = function(url, filterAttributeMap) {
		
		if(typeof url != "string")
			throw new Error("CollectionDataSource: url must be a string");
			
		var _qsArgs = [];
		var cdsObj = {};
		
		var map = function(x) {
			if(filterAttributeMap == null) return x;
			return filterAttributeMap.map(x);
		};
		
		var resultMapperFunc;
		cdsObj.setResultMapper = function(mapperFunc) {
			if(typeof mapperFunc !== "function")
				throw new Error("setResultMapper must be passed a function");
				
			resultMapperFunc = mapperFunc;
		};
		
		cdsObj.stream = new Rx.Subject();	
		cdsObj.errorStream = new Rx.Subject();
		
		cdsObj.addQueryStringArg = function(name, value) {
			_qsArgs.push({name: name, value: value});
		};
		
		cdsObj.getOptionsObject = getOptionsObject;
		
		// Normally the success stream doesn't get anything on a failure, but that would
		// keep the CompoundDataSource from working if any sources in it fail.  This flag
		// allows us to emit nulls on the success stream, and allows CompoundDataSource
		// to aggregate the successful streams.
		cdsObj.emitNullValueOnFailure = false;
		
		cdsObj.query = function(options) {
			if(typeof options !== "object" || options === null) throw new Error("expected an object");
			var skip =  typeof options.skip === "number" && options.skip >= 0
				? options.skip : 0;
			var top = typeof options.top === "number" 
				? options.top : 10;
			var orderBy = typeof options.orderBy === "string" && options.orderBy != ""
				? map(options.orderBy) : null;
			var orderAsc = typeof options.orderAsc === "boolean" && options.orderAsc === true
				? " asc" : " desc"
				
			var arrQs = [];
			
			for(var i=0; i<_qsArgs.length; i++) {
				arrQs.push(_qsArgs[i].name + "=" + _qsArgs[i].value);
			}
			arrQs.push("skip=" + skip);
			arrQs.push("top=" + top);	
			var sourceStore = options.sourceStore;
			if(sourceStore != null){
				if(typeof sourceStore.name == "string" && sourceStore.name != "") {
					arrQs.push("sourcestore=" + sourceStore.name);
				}
			}			
			
			if(typeof options == "object" && options != null && typeof options.filter === "object" && options.filter.length > 0) {
				var filtersArray = [];
				for(var i = 0; i<options.filter.length; i++) {
					var filter = options.filter[i];
					var prepedValue = typeof filter.value === "string" ? "'" + filter.value + "'" : filter.value;
					filtersArray.push(map(filter.attribute) + " " + filter.operator + " " + encodeURIComponent(prepedValue));
				}
				arrQs.push("filter=" + filtersArray.join(" AND "));
			}	
			
			if(orderBy != null)
				arrQs.push("orderBy=" + orderBy + orderAsc);
			
			if(options.attributeInfoOnly)
				arrQs.push("attributeInfoOnly=1");
			
			var qs = arrQs.join("&");
			var promise = $.getJSON(PortalTools.urlPrefix + url + "?" + qs);
			promise.then(function(result) {
				try {
					var mappedData = resultMapperFunc != null ? 
						_(result.data).map(resultMapperFunc) :
						result.data;
					result.data = mappedData;
					cdsObj.stream.onNext(result);
				} catch(e) {
					handleFailure(null);
				}
			}).fail(handleFailure);
			
			function handleFailure(data) {
				cdsObj.errorStream.onNext(data);
				if(cdsObj.emitNullValueOnFailure) {
					cdsObj.stream.onNext(null);
				}
			}
		};

		return cdsObj;
	};
	
	var CompoundDataSource = function(numPerSource) {
		var _dataSources = [];
		var _numPerSource = numPerSource;
		
		var cdsObj = {};
		
		cdsObj.stream = new Rx.Subject();
		cdsObj.errorStream = new Rx.Subject();
		
		cdsObj.addDataSource = function(dataSource, mapFunction) {
			_dataSources.push({source: dataSource, map: mapFunction});
		};
		
		cdsObj.getOptionsObject = getOptionsObject;
		
		cdsObj.query = function(options) {
			// we can never use skip in the child streams because paging
			// might get screwed up, so always get everything up to the
			// amount of data requested.
			var subOptions = {
				skip: 0,
				top: options.skip + options.top,
				orderBy: options.orderBy,
				orderAsc: options.orderAsc,
				filter: options.filter
			};
			
			var sources = _(_dataSources).pluck("source");
			
			// pass along any errors from any of the streams to the subscribers
			// of this CompoundDataSource's errorStream
			_(sources).each(function(src) {
				// if this is not set, the success handler will not run to combine the sources
				// that do succeed in loading their data
				src.emitNullValueOnFailure = true;
				
				// pass along the individual errors to the errorStream, rather than aggregating
				// them, to ensure useful data isn't lost
				src.errorStream.subscribe(function(data) {
					cdsObj.errorStream.onNext(data);
				});
			});
			
			var streams = _(sources).pluck("stream");
			var bigStream = Rx.Observable.combineLatest(streams);
			bigStream.subscribe(function(subResults) {
				
				// Any streams that suffered exceptions will return nulls, so we should filter
				// those out and just handle the successful requests.  Seperate failure handlers
				// should be used to hande the failures.
				var validResults = _(subResults).filter(function(x) {
					return x !== null;
				});
			
				var names = _(validResults).pluck("collectionName");
				var types = _.chain(validResults).pluck("collectionType").uniq(false).value();
				var arrEachNumberInSet = validResults.map(function(x) {return x.numberInSet});
				var combinedNumberInSet = 
					arrEachNumberInSet.some(function(x) {return x === -1;}) ? -1 :
					arrEachNumberInSet.reduce(function(x, y) { return x+y; }, 0);
				var mergedData = _.chain(validResults).pluck("data").flatten(true).value();
				var sortedData = mergedData = typeof options.orderBy === "string" && options.orderBy != "" ?
					_(mergedData).sortBy(options.orderBy) :
					mergedData; // no sorting done
				var finalSortedData = options.orderAsc ? sortedData : sortedData.reverse();
				var finalData = _.chain(finalSortedData).rest(options.skip).first(options.top).value();
				var result = {
					collectionName: "Compound Data Source: " + names.join(", "),
					previousPageUrl: null,
					nextPageUrl: null,
					numberInSet: combinedNumberInSet,
					collectionType: types.length == 1 ? types[0] : types.join(", "),
					data: finalData
				};
				cdsObj.stream.onNext(result);
			});
			Rx.Observable.from(sources).subscribe(function(s) {
				s.query(subOptions);
			});
		};
		
		return cdsObj;
	};
	
	var FilterAttributeMap = function() {
		var map = {};
		var publicInterface = {};
		publicInterface.add = function(x, y) {
			map[x] = y;
		};
		publicInterface.map = function(x) {
			var z = map[x];
			return typeof z == "undefined" ? x : z; 
		};
		return publicInterface;
	};
	
	PortalTools.ApiDataSource = {};
	
	PortalTools.ApiDataSource.CustomSearchIndex = function(arrAttributes, filterAttributeMap) {
		return CollectionDataSource("codex/savedsearch", arrAttributes, filterAttributeMap);
	}
	
	PortalTools.ApiDataSource.GetCustomSearchByPORef = function(poref, arrAttributes, filterAttributeMap) {
		var hexStr = poref
			.replace("com.webridge.entity.Entity[OID[", "")
			.replace("]]", "");
		var dataSource = CollectionDataSource("codex/savedsearch/" + hexStr, arrAttributes, filterAttributeMap);
		
		dataSource.addQueryStringArg("search", hexStr);
		return dataSource;
	};
	
	PortalTools.ApiDataSource.GetEntitySetWebApi = function(apiType, arrAttributes, filterAttributeMap) {
		var dataSource = CollectionDataSource("codex/entitysetwebapi/" + apiType, arrAttributes, filterAttributeMap);
		dataSource.addQueryStringArg("etype", apiType);
		return dataSource;
	};
	
	PortalTools.CollectionDataSource = CollectionDataSource;
	PortalTools.CompoundDataSource = CompoundDataSource;
	PortalTools.FilterAttributeMap = FilterAttributeMap;
}());